#!/usr/bin/perl -w

use strict;
use Socket;

$ENV{PATH} = '/usr/local/bin:/usr/bin:/bin';
$| ++;


# Inladen plugins...
my $sntmondir = '/etc/snt-mon/';
#$sntmondir = '../';

chdir $sntmondir
	or die "Could not chdir to $sntmondir: $!\n";

my %plugins = (); # name => [ sub, interval, last ]

if (opendir DIR, 'plugins/') {
	while (my $name = readdir DIR) {
		next if $name =~ /^\./;
		next if $name =~ /\.dpkg-/;
		next if $name =~ /\.old$/;
		next if $name =~ /~$/;
		my $file = "plugins/$name";

		if ( -f $file && -x $file ) {
			if (open PLUGIN, "< $file") {
				my $func = '';
				my $interval = 300;

				foreach my $line (<PLUGIN>) {
					chomp $line;
					# '#@ Interval = 30'
					if ($line =~ /^#\@\s+interval\s*=\s*([1-9]\d*)\s*$/i) {
						$interval = $1;
					} else {
						$line =~ s/\bprint\b/ push \@lines, /g;
						$func .= $line . "\n";
					}
				}

				$func = eval "sub { my \@lines; $func; return \@lines }";
				die "Could not load plugin '$name':\n$@" if $@;

				$plugins{$name} = [ $func, $interval, 0 ];
			} else {
				die "Could not open plugin '$name': $!\n";
			}
		}
	}
} else {
	die "Missing plugins dir: $!\n";
}



# Daemon...
my $pidfile = '/var/run/snt-mon.pid';
#$pidfile = '/tmp/snt-mon.pid';

die "Pid-file exists, daemon already running!\n"
	if -f $pidfile;

my $pid = fork();
if (! defined $pid) {
	die "Failed to fork: $!\n";

} elsif ($pid ne '0') {
	if (open PID, "> $pidfile") {
		print PID "$pid\n";
		close PID;
	} else {
		die "Daemon forked, but couldn't write pid-file.\n";
	}

	exit(0);
}

chdir '/';
close STDIN; open STDIN, "< /dev/null";
close STDOUT; open STDOUT, "> /dev/null";
close STDERR; open STDERR, "> /dev/null";

$SIG{'HUP'} = $SIG{'TERM'} = sub {
		unlink $pidfile;
		exit(0);
	};



# Syslog...
my $host = 'dropveter.snt.utwente.nl';

my $logprio = 14; # user.info = (LOG_USER (=1<<3) | LOG_INFO (=6))
my $hostname = `hostname`;
chomp $hostname;
$hostname =~ s/\..*$//g;

my $host_addr = undef;
while (!defined $host_addr) {
	$host_addr = inet_aton($host) or undef;
	if (!defined $host_addr) {
		warn "inet_aton: $!\n";
		sleep 300;
	}
}

socket(my $sock, PF_INET, SOCK_DGRAM, getprotobyname('udp'))
	or warn "socket: $!\n";

my $destpaddr = sockaddr_in(514, $host_addr) or warn "sockaddr_in: $!\n";



# running..
my %seen = ();

while (1) {
	my $ts = time;
	my $next = $ts + 300;

	foreach my $p (values %plugins) {
		my ($func, $interval, $last) = @$p;

		# time-shift protection.. :-/
		if ($last > $ts) {
			warn "Time-shift detected!\n";
			$last = $ts;
		}

		# time to run plugin?
		if ($last + $interval <= $ts) {
			# syslog time format
			my $time = substr(scalar localtime, 4, 15);
			my @lines = &$func;
			foreach my $line (@lines) {
				chomp $line;
				if ($line =~ /^([^:"]*\s*("([^\\"]*|\\[\\"])*")?):(.*)$/s) {
					my ($key, $val) = ($1, $4);
					if (!$seen{$key}) {
						$seen{$key} = 1;
						$val =~ s/=\d+/=U/g;
						$line = $key . ':' . $val;
					}
					$line = "<$logprio>$time $hostname STATS: [$ts] $line\n";
					send($sock, $line, 0, $destpaddr)
						or warn "send: $!\n";
				}
			}

			while ($last + $interval <= $ts) {
				$last += $interval;
			}
		}

		$next = $last + $interval
			if $next > $last + $interval;

		$p->[2] = $last;
	}

	$next -= time;
	sleep $next if $next > 0;
}
